"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from tributors.utils.command import Command
import logging
import os
import re
import requests
import sys

repository_regex = "(?P<owner>[\w,\-,\_]+)/(?P<repo>[\w,\-,\_\.]+)"

bot = logging.getLogger("github")


class GitHubRepository:
    """A GitHub repository parses a repo and exposes repository and contributor
       metadata.
    """

    def __init__(self, repo, skip_users=None, params=None):
        self._repo = {}
        self._contributors = []
        self._topics = []
        self.uid = get_github_repository(repo)
        self.skip_users = skip_users or []
        self.params = params or {}
        self.cache = {}

    def include_contributor(self, login):
        """Given a threshold (and preference to not include bots) return a boolean
           to indicate including the contributor or not
        """
        contributor = self.contributors.get(login)

        # if the login is marked to skip
        if login in self.skip_users:
            return False

        # Skip GitHub bots
        if contributor["type"] == "Bot" or "[bot]" in contributor["login"]:
            return False
        return True

    # GitHub repository can serve as a metadata parser
    def update_lookup(self):
        """update the .tributors file using GitHub contributors
        """
        if not self.skip_users:
            self.skip_users = self.params.get("--skip-users", "").split(" ")

        # Iterate through GitHub and update metadata
        for login, contributor in self.contributors.items():

            # Don't include bots, and others specified with --skip-user
            if not self.include_contributor(login):
                continue

            # Look up a GitHub username, possibly email and site
            user = get_user(login)

            entry = {"name": user.get("name") or login}
            if login in self.cache:
                entry = self.cache[login]
            else:
                bot.info(f"⭐️ new contributor {login}")

            # Update cache with fields that aren't defined yet
            for key in ["email", "bio", "blog"]:
                if user.get(key) and key not in entry:
                    entry[key] = user.get(key)
                    bot.info(f"  Updating {login} with {key}: {entry[key]}")
            self.cache[login] = entry
        return self.cache

    def __str__(self):
        return "[github][%s]" % self.uid

    def __repr__(self):
        return self.__str__()

    @property
    def repo(self):
        """Retrieve the GitHub repository, if we don't have it yet
        """
        if not self._repo:
            self._repo = get_repo(self.uid)
        return self._repo

    # Equivalent methods to a parser to get lookups based on primary ids
    @property
    def email_lookup(self):
        """Return loaded metadata as an email lookup.
        """
        return {}

    @property
    def name_lookup(self):
        """Return loaded metadata as an orcid lookup.
        """
        return {}

    @property
    def orcid_lookup(self):
        """Return loaded metadata as an orcid lookup.
        """
        return {}

    @property
    def login_lookup(self):
        """Return loaded metadata as a github login lookup.
        """
        return self.contributors

    @property
    def contributors(self):
        """Return list of contributors, and retrieve if we don't have yet
        """
        if not self._contributors:
            self._contributors = get_contributors(self.uid)
        return self._contributors

    def topics(self, topics=None):
        """Return list of topics, optionally add extras and return unique set
        """
        if not self._topics:
            self._topics = get_topics(self.uid)
        topics = topics or []
        return list(set(self._topics + topics))

    @property
    def description(self):
        return self.repo["description"]

    @property
    def html_url(self):
        return self.repo["html_url"]

    @property
    def name(self):
        return self.repo["name"]

    @property
    def issues_url(self):
        return "%s/issues" % self.repo["html_url"]

    @property
    def license(self):
        return "https://spdx.org/licenses/%s" % self.repo["license"]["spdx_id"]


def get_topics(repo):
    """Given a repository, get topics associated.
    """
    repo = get_repo(repo) or {}
    return repo.get("topics", [])


def get_repo(repo):
    """get_repo will return a single repo, username/reponame
       given authentication with user    
    """
    headers = get_headers()
    headers["Accept"] = "application/vnd.github.mercy-preview+json"
    url = "https://api.github.com/repos/%s" % repo
    response = requests.get(url, headers=headers)

    # Case 2: public and private
    if response.status_code != 200:
        bot.warning(f"Issue getting {repo}: {response.status_code}, {response.reason}")
        return

    return response.json()


def get_contributors(repo):
    """Given a GitHub repository address, retrieve a lookup of contributors
       from the API endpoint. We look to use the GITHUB_TOKEN if exported
       to the environment, and exit if the response has any issue.
    """
    if not repo:
        sys.exit("A repository is required to get contributors.")
    url = "https://api.github.com/repos/%s/contributors" % repo
    headers = get_headers()
    response = requests.get(url, headers=headers)
    if response.status_code != 200:
        message = "Response %s from GitHub: %s, cannot retrieve contributors " % (
            response.status_code,
            response.reason,
        )
        if not os.environ.get("GITHUB_TOKEN"):
            message += " you should export GITHUB_TOKEN to increase your API limits"
        sys.exit(message)

    # Return a lookup based on GitHub Login
    return {x["login"]: x for x in response.json()}


def get_headers():
    """Get headers, including a Github token if found in the environment
    """
    token = os.environ.get("GITHUB_TOKEN")
    headers = {"Accept": "application/vnd.github.v3+json"}
    if token:
        headers["Authorization"] = "token %s" % token
    return headers


def get_user(username):
    """Given a username, retrieve the user metadata from GitHub. We need to do
       this to get the profile (blog) url from the metadata
    """
    url = "https://api.github.com/users/%s" % username
    headers = get_headers()
    response = requests.get(url, headers=headers)
    if response.status_code != 200:
        sys.exit(
            "Response %s: %s, cannot retrieve GitHub user %s."
            % (response.status_code, response.reason, username)
        )
    return response.json()


def get_github_repository(repo):
    """First preference goes to repo variable provided, then check the environment,
       and then check for a local .git repo. Finally, verify that format is 
       correct. Return the repository name.
    """
    if isinstance(repo, GitHubRepository):
        return repo.uid
    repo = repo or os.environ.get("GITHUB_REPOSITORY")
    if not repo:
        command = Command("git config --get remote.origin.url")
        command.execute()

        # Issue running command
        if command.returncode != 0 or not command.out:
            sys.exit("Could not determine repository from local .git.")
        repo = "/".join(command.out[0].strip().split("/")[-2:])

    match = re.search(repository_regex, repo)

    # Must conform to correct name
    if not match:
        sys.exit("Malformed repository address %s" % repo)
    owner, repo = match.groups()
    return "%s/%s" % (owner, re.sub(".git$", "", repo, 1))



#!/usr/bin/env python3

"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

import logging
import os
import sys

from tributors.main.github import GitHubRepository
from tributors.utils.file import write_json
from .base import ParserBase

bot = logging.getLogger("allcontrib")


class AllContribParser(ParserBase):

    name = "allcontrib"

    # https://allcontributors.org/docs/en/emoji-key
    contribution_types = [
        "audio",
        "ally",
        "bug",
        "blog",
        "business",
        "code",
        "content",
        "data",
        "doc",
        "design",
        "example",
        "eventOrganizing",
        "financial",
        "fundingFinding",
        "ideas",
        "infra",
        "maintenance",
        "platform",
        "plugin",
        "projectManagement",
        "question",
        "review",
        "security",
        "tool",
        "translation",
        "test",
        "tutorial",
        "talk",
        "userTesting",
        "video",
    ]

    def __init__(self, filename=None, repo=None, params=None):
        filename = filename or ".all-contributorsrc"
        self.extras = []
        super().__init__(filename, repo, params)

    def load_data(self):
        return self._load_data("--allcontrib-file")

    @property
    def login_lookup(self):
        self.load_data()
        return {
            x["login"]: x for x in self.data.get("contributors", []) if "login" in x
        }

    @property
    def name_lookup(self):
        """Return loaded metadata as an name lookup.
        """
        self.load_data()
        return {x["name"]: x for x in self.data.get("contributors", []) if "name" in x}

    def init(self, force=False, from_resources=None, save=True):
        """Given an allcontributors file (we default to the one expected) and
           a preference to force, write the empty file to the repository.
           If the file exists and force is false, exit on error. If the user
           has not provided a full repository name and it's not in the environment,
           also exit on error

           Arguments:
            - repo (str)     : the full name of the repository on GitHub
            - force (bool)   : if the contributors file exists, overwrite
            - filename (str) : default filename to write to.
        """
        filename = self.params.get("--allcontrib-file", self.filename)
        if os.path.exists(filename) and not force:
            sys.exit("%s exists, set --force to overwrite." % filename)

        bot.info(f"Generating {filename} for {self.repo.uid}")
        owner, repo = self.repo.uid.split("/")[:2]

        # Write metadata to empty all contributors file.
        metadata = {
            "projectName": repo,
            "projectOwner": owner,
            "repoType": "github",
            "repoHost": "https://github.com",
            "files": ["README.md"],
            "imageSize": 100,
            "commit": True,
            "commitConvention": "none",
            "contributors": [],
            "contributorsPerLine": 7,
        }
        if save:
            write_json(metadata, filename)
        return metadata

    def update(self, thresh=1, from_resources=None, save=True):
        """Given an existing contributors file, use the GitHub API to retrieve
           all contributors, and then use subprocess to update the file
        """
        self.thresh = thresh
        self.load_data()
        from_resources = from_resources or {}

        bot.info(f"Updating {self.filename}")

        # Get optional (or default) contributor type
        ctype = self.params.get("--allcontrib-type", "code")
        if ctype not in self.contribution_types:
            sys.exit(
                f"Invalid contribution type {ctype}. See https://allcontributors.org/docs/en/emoji-key for types."
            )

        # Sanity check that we have the correct repository
        repo = "%s/%s" % (self.data["projectOwner"], self.data["projectName"])

        if repo != self.repo.uid:
            bot.warning(
                f"Found different repository {repo} in {self.filename}, updating from {self.repo.uid}"
            )
            self._repo = GitHubRepository(repo)

        # Update the cache from GitHub, and .tributors lookup
        self.update_cache()

        # Parse over logins, start with existing values
        self.lookup = {
            x["login"]: x for x in self.data.get("contributors", []) if "login" in x
        }
        self.extras = [x for x in self.data.get("contributors", []) if "login" not in x]

        # This client only supports update from names and logins
        self.update_from_logins(from_resources.get("login", []), ctype)
        self.update_from_names(from_resources.get("names", []), ctype)
        self.update_from_orcids(from_resources.get("orcid", []))
        self.update_from_emails(from_resources.get("email", []))

        # Update the contributors
        self.data["contributors"] = list(self.lookup.values()) + self.extras

        if save:
            write_json(self.data, self.filename)
        return self.data

    def update_from_names(self, names, ctype):
        """Given a list of names, update the loaded logins
        """
        for name in names:
            if name not in self.name_lookup:
                bot.info(f"⭐️ Found new contributor {name} in {self.filename}")
                entry = {
                    "name": name,
                    "contributions": [ctype],
                }
                self.extras.append(entry)

    def update_from_logins(self, logins, ctype):
        """Given a list of logins, update the loaded logins
        """
        for login in logins:

            # Check against contribution threshold, and not bot
            if not self.include_contributor(login):
                continue

            cache = self.cache.get(login) or {}
            if login in self.login_lookup:
                entry = self.login_lookup[login]
            else:
                bot.info(f"⭐️ Found new contributor {login} in {self.filename}")
                entry = {
                    "login": login,
                    "name": cache.get("name") or login,
                    "contributions": [ctype],
                }

            # Only add profile and profile if not added yet
            if "profile" not in entry:
                entry["profile"] = cache.get("blog") or self.repo.contributors.get(
                    login, {}
                ).get("html_url")
            if "avatar_url" not in entry:
                entry["avatar_url"] = (
                    self.repo.contributors.get(login, {}).get("avatar_url"),
                )

            if ctype not in entry["contributions"]:
                entry["contributions"].append(ctype)
            self.lookup[login] = entry

    def update_lookup(self):
        """Each client optionally has it's own function to update the cache.
           In the case of allcontributors, we run this function on update after
           self.lookup is defined with current data. We use this lookup to
           update a shared cache that might be used for other clients. Since
           we also have self.contributors (with GitHub responses) we don't need
           to add items that would be found there.
        """
        self.load_data()
        bot.info(f"Updating .tributors cache from {self.filename}")
        for login, metadata in self.login_lookup.items():
            if login in self.cache:
                entry = self.cache[login]
            else:
                entry = {}
                bot.info(f"⭐️ Found new contributor {login} in {self.filename}")
            if "name" not in entry and "name" in metadata:
                entry["name"] = metadata["name"]
                bot.info(f"   Updating {login} with name: {entry['name']}")
            if "blog" not in entry and "profile" in metadata:
                entry["blog"] = metadata["profile"]
                bot.info(f"   Updating {login} with blog: {entry['blog']}")
            self.cache[login] = entry



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

import json
import os
import tempfile


def write_json(json_obj, filename, pretty=True):
    """write_json will write a json object to file, pretty printed

       Arguments:
        - json_obj (dict) : the dict to print to json
        - filename (str)  : the output file to write to
    """
    with open(filename, "w", encoding="utf8") as filey:
        if pretty:
            filey.writelines(
                json.dumps(
                    json_obj, indent=4, ensure_ascii=False, separators=(",", ": ")
                )
            )
        else:
            filey.writelines(json.dumps(json_obj), ensure_ascii=False)
    return filename


def read_json(input_file):
    """Read json from an input file.

       Arguments:
         - input_file (str) : the filename to read
    """
    with open(input_file, "r") as filey:
        data = json.loads(filey.read())
    return data


def read_file(filename, readlines=True):
    """write_file will open a file, "filename" and write content
       and properly close the file.

       Arguments:
         - filename (str) : the filename to read
         - readlines (bool) : read lines of the file (vs all raw)
    """
    with open(filename, "r") as filey:
        if readlines is True:
            content = filey.readlines()
        else:
            content = filey.read()
    return content


def write_file(filename, content):
    """write some content to a filename"""
    with open(filename, "w") as fd:
        fd.writelines(content)
    return filename


def get_tmpfile(prefix="tributors-"):
    """get a temporary file with an optional prefix. By default, the file
       is closed (and just a name returned).

       Arguments:
        - prefix (str) : prefix with this string
    """
    tmpdir = tempfile.gettempdir()
    prefix = os.path.join(tmpdir, os.path.basename(prefix))
    fd, tmp_file = tempfile.mkstemp(prefix=prefix)
    os.close(fd)
    return tmp_file



import os

LOG_LEVEL = os.environ.get("TRIBUTORS_LOG_LEVEL", "INFO")
LOG_LEVELS = ["DEBUG", "CRITICAL", "ERROR", "WARNING", "INFO", "QUIET", "FATAL"]
if LOG_LEVEL not in LOG_LEVELS:
    LOG_LEVEL = "INFO"



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from tributors.main.parsers import get_named_parser
from tributors.utils.file import write_json, read_json
from .github import GitHubRepository
import logging
import os

bot = logging.getLogger("tributors.main")


class TributorsClient:
    """The tributors client is the handler to interact with one or more
       contributor actions. If we do an update for multiple, for example,
       we can cache and re-use the GitHub calls.
    """

    def __init__(self, skip_cache=False):
        """create a tributors client to control one or more updates to
           contribution files. The .tributors cache stores identifiers that
           would need to be looked up, and the client stores a contributors
           cache (from GitHub) that can be used between parser clients.
        """
        if not skip_cache:
            self.load_cache()
        self.skip_cache = skip_cache

    def load_cache(self):
        """load a cache to serve as a lookup for contributors.
           Each parser will use the cache to find common identifiers,
           and update it if necessary. We use the cache as a place to
           store emails / orcid id / username combos. For temporary 
           (GitHub request) caches, we use /tmp.
        """
        self.cache = {}
        if os.path.exists(".tributors"):
            self.cache = read_json(".tributors")

    def save_cache(self):
        """Save the current self.cache to the cache file .tributors in the PWD
        """
        if not self.skip_cache:
            bot.debug("Saving cache to .tributors")
            write_json(self.cache, ".tributors")

    def __str__(self):
        return "[TributorsClient]"

    def __repr__(self):
        return self.__str__()

    def init(
        self,
        parsers=None,
        repo=None,
        params=None,
        force=False,
        skip_users=None,
        from_resources=None,
    ):
        """Init one or more contributor parsers. Specifically, this is the
           action that runs the parser.init() to generate some initial file.
        """
        parsers = parsers or []

        # Generate a shared repository object
        repo = GitHubRepository(repo, skip_users)

        # Get resource lookup ids (emails, orcids, logins, most won't be used for init)
        resources = self.get_resource_lookups(from_resources, params)

        for parser in parsers:
            client = get_named_parser(name=parser, repo=repo, params=params)
            client.cache = self.cache
            client.init(force=force, from_resources=resources)
            self.cache.update(client.cache)

        # Save the cache
        self.save_cache()

    def update_resource(self, resources=None, params=None, skip_users=None):
        """Given one or more resource types (an external file or source of
           metadata) update the .tributors cache lookup
        """
        resources = resources or []
        for name in resources:
            resource = get_named_parser(name=name, params=params)
            resource.cache = self.cache
            resource.update_lookup()

        # Save the cache
        self.save_cache()

    def update(
        self,
        parsers=None,
        repo=None,
        params=None,
        thresh=1,
        skip_users=None,
        from_resources=None,
    ):
        """Update one or more contributor parsers. Specifically, this is the
           action that runs the parser.update() after obtaining contributions
           from GitHub or a cache.
        """
        parsers = parsers or []

        # Generate a shared repository object
        repo = GitHubRepository(repo, skip_users=skip_users)
        bot.debug(f"Found repository {repo}")

        # Get resource lookup ids (emails, orcids, logins)
        resources = self.get_resource_lookups(from_resources, params)

        # Update each metadata file via its parser
        for parser in parsers:
            client = get_named_parser(name=parser, repo=repo, params=params)
            client.cache = self.cache
            client.update(thresh=thresh, from_resources=resources)
            self.cache.update(client.cache)

        # Save the cache
        self.save_cache()

    def get_resource_lookups(self, from_resources=None, params=None):
        """Based on a name (e.g., GitHub or special case tributors) return 
           as many lookups of unique ids (email, login, orcid) that
           the resource provides
        """
        from_resources = from_resources or ["github"]
        lookups = {"login": set(), "orcid": set(), "email": set(), "name": set()}

        for name in from_resources:

            # Special case, tributors is just the entire cache
            if name == "tributors":
                lookups["login"].update(self.cache)
                lookups["name"].update(
                    {x["name"]: x for _, x in self.cache.items() if "name" in x}
                )
                lookups["orcid"].update(
                    {x["orcid"]: x for _, x in self.cache.items() if "orcid" in x}
                )
                lookups["email"].update(
                    {x["email"]: x for _, x in self.cache.items() if "email" in x}
                )
            else:
                parser = get_named_parser(name=name, params=params)
                lookups["name"].update(parser.name_lookup)
                lookups["login"].update(parser.login_lookup)
                lookups["email"].update(parser.email_lookup)
                lookups["orcid"].update(parser.orcid_lookup)
        return lookups



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from tributors.main.github import GitHubRepository
from .allcontrib import AllContribParser
from .codemeta import CodeMetaParser
from .zenodo import ZenodoParser
from .mailmap import MailmapParser
import re


def get_named_parser(name, repo=None, filename=None, params=None):
    """get a named parser, meaning determining based on name and not uri.
       Typically a parser is used to update a metadata file and also
       update the shared .tributors cache, but in a few cases (where we don't
       have any metadata file to update) the parser is just used as a resource
       to update the cache.
    """
    if repo and not isinstance(repo, GitHubRepository):
        repo = GitHubRepository(repo)
    parser = None

    # These parsers have contributors files, and used to update .tributors
    if re.search("(allcontrib|all-contrib)", name):
        parser = AllContribParser(filename, repo, params)
    elif re.search("zenodo", name):
        parser = ZenodoParser(filename, repo, params)
    elif re.search("codemeta", name):
        parser = CodeMetaParser(filename, repo, params)

    # These parsers only update .tributors metadata
    elif re.search("mailmap", name):
        parser = MailmapParser(filename=filename, params=params)
    elif re.search("github", name):
        parser = GitHubRepository(repo=repo, params=params)

    if not parser:
        raise NotImplementedError(f"There is no matching parser for {name}")

    return parser



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from tributors.main import TributorsClient
from .utils import parse_extra
import os
import sys


def main(args, extra):

    client = TributorsClient(skip_cache=args.skip_cache)

    # Parse extra arguments
    extra = parse_extra(extra)
    extra["--interactive"] = args.interactive

    # Start with user provided parsers
    parsers = args.parsers

    # Skip users
    skip_users = []
    if args.skip_users != "unset":
        skip_users = args.skip_users

    # Does the user want to update from a particular resource?
    from_resources = args.from_resources
    if from_resources == "unset":
        from_resources = None

    # If unset, try to detect files
    if "unset" in parsers:
        lookup = {
            "allcontrib": extra.get("--allcontrib-file", ".all-contributorsrc"),
            "zenodo": extra.get("--zenodo-file", ".zenodo.json"),
            "codemeta": extra.get("--codemeta-file", "codemeta.json"),
        }

        parsers = []
        for parser, filename in lookup.items():
            if os.path.exists(filename):
                parsers.append(parser)

        # Exit if no parsers auto-detected
        if not parsers:
            sys.exit("No parsers auto-detected. Specify a parser name instead?")

    if "all" in parsers:
        client.update(
            parsers=["zenodo", "allcontrib", "codemeta"],
            thresh=args.thresh,
            repo=args.repo,
            params=extra,
            skip_users=skip_users,
            from_resources=from_resources,
        )

    else:
        client.update(
            parsers=parsers,
            repo=args.repo,
            params=extra,
            thresh=args.thresh,
            skip_users=skip_users,
            from_resources=from_resources,
        )



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from tributors.main.orcid import get_orcid, OrcidIdentifier
from tributors.utils.file import read_json

import logging
import os
import sys

bot = logging.getLogger("tributors")


class ParserBase:
    """A parser base exists to provide structure to init and update contributor
       configuration files.
    """

    name = "base"

    def __init__(self, filename=None, repo=None, params=None):
        """initialize a new contributor parser.
        """
        self.filename = filename
        self._repo = repo
        self.cache = {}
        self.contributors = {}
        self.thresh = 1
        self.params = params or {}
        self.data = {}

    def __str__(self):
        if self.filename:
            return "[%s:%s]" % (self.name, self.filename)
        return "[%s]" % (self.name)

    def __repr__(self):
        return self.__str__()

    # Methods that parsers expose to get contributors

    @property
    def email_lookup(self):
        """Return loaded metadata as an email lookup.
        """
        return {}

    @property
    def orcid_lookup(self):
        """Return loaded metadata as an orcid lookup.
        """
        return {}

    @property
    def name_lookup(self):
        """Return loaded metadata as an name lookup.
        """
        return {}

    @property
    def login_lookup(self):
        """Return loaded metadata as a github login lookup.
        """
        return {}

    @property
    def repo(self):
        """after some initial parsing, we can retrieve the name of the GitHub repo
        """
        return self._repo

    def update_from_emails(self, *args, **kwargs):
        if args:
            bot.warning(f"{self.name} does not support updating from email.")

    def update_from_logins(self, *args, **kwargs):
        if args:
            bot.warning(f"{self.name} does not support updating from logins.")

    def update_from_orcids(self, *args, **kwargs):
        if args:
            bot.warning(f"{self.name} does not support updating from orcids.")

    def update_from_names(self, *args, **kwargs):
        if args:
            bot.warning(f"{self.name} does not support updating from names.")

    def init(self, *args, **kwargs):
        """init a new configuration file
        """
        raise NotImplementedError

    def update(self, *args, **kwargs):
        """update a configuration file
        """
        raise NotImplementedError

    def _load_data(self, fileattr):
        """Load self.filename unless a file attribute is defined
        """
        if not self.data:
            filename = self.params.get(fileattr, self.filename)

            if not os.path.exists(filename):
                sys.exit("%s does not exist" % filename)

            self.data = read_json(filename)
            self.filename = filename
        return self.data

    def include_contributor(self, login):
        """Given a threshold (and preference to not include bots) return a boolean
           to indicate including the contributor or not
        """
        contributor = self.repo.contributors.get(login)

        # if the login is marked to skip
        if login in self.repo.skip_users:
            return False

        # If they don't meet the threshold, continue
        if contributor and contributor["contributions"] < self.thresh:
            return False

        # Skip GitHub bots
        if (
            contributor
            and contributor["type"] == "Bot"
            or contributor
            and "[bot]" in contributor["login"]
        ):
            return False
        return True

    def update_cache(self, update_lookup=True):
        """A shared function to run additional parsing on the cache, such
           as adding an orcid id when an email is defined. 
        """
        interactive = self.params.get("--interactive", False)

        # If the parser can be used as a resource, use it to update .tributors
        if hasattr(self, "update_lookup") and update_lookup:
            self.update_lookup()

        # Then add an Orcid lookup
        for login, entry in self.cache.items():

            # If we have an email, and orcid isn't defined
            if "orcid" not in entry:
                orcid = get_orcid(
                    entry.get("email"), entry.get("name"), interactive=interactive
                )
                if orcid:
                    entry["orcid"] = orcid
                    cli = OrcidIdentifier(orcid)

                    # If we found the record, update metadata
                    if (
                        cli.found
                        and not entry.get("name")
                        or entry.get("name") == login
                    ):
                        entry["name"] = "%s %s" % (cli.firstName, cli.lastName)
                    if cli.affiliation and not entry.get("affiliation"):
                        entry["affiliation"] = cli.affiliation

            self.cache[login] = entry



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

import logging
import os
import sys

from tributors.utils.file import read_file
from .base import ParserBase

bot = logging.getLogger("   mailmap")


class MailmapParser(ParserBase):

    name = "mailmap"

    def __init__(self, filename=None, params=None, **kwargs):
        filename = filename or ".mailmap"
        super().__init__(filename=filename, params=params)

    def load_data(self):
        """Since mailmap has format Name <email> on each line, we have a custom
           loading function, and we don't ever need to write anything
        """
        if not self.data:
            self.filename = self.params.get("--mailmap-file", self.filename)

            # Ensure codemeta file already exists
            if not os.path.exists(self.filename):
                sys.exit("%s does not exist" % self.filename)

            for line in read_file(self.filename):
                name, email = line.split("<")
                email = email.strip().rstrip(">")
                self.data[email] = {"name": name.strip()}
        return self.data

    @property
    def email_lookup(self):
        """Return loaded metadata as an email lookup. In this case, this
           is just the entire data.
        """
        if not hasattr(self, "_email_lookup"):
            self.load_data()
            self._email_lookup = self.data
        return self._email_lookup

    def update_lookup(self):
        """Each client optionally has it's own function to update the cache.
            In the case of zenodo, we aren't necessarily aware of GitHub
            login (the current mapping mechanism) so we cannot update the
            cache yet. When orcid is added this might be an option.
        """
        bot.info(f"Updating .tributors cache from {self.filename}")
        self.load_data()

        # Find matches based on email
        for login, cache in self.cache.items():
            email = cache.get("email")
            if email in self.email_lookup:
                if "name" not in cache:
                    cache["name"] = self.email_lookup[email]["name"]
                    bot.info(f"   Updating {login} with name: {cache['name']}")



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from tributors.main import TributorsClient
from .utils import parse_extra
import logging
import sys

bot = logging.getLogger("github")


def main(args, extra):

    client = TributorsClient(skip_cache=args.skip_cache)

    # Parse extra arguments
    extra = parse_extra(extra)
    extra["--interactive"] = args.interactive

    # Skip users, if a space separated list is defined
    skip_users = []
    if args.skip_users != "unset":
        skip_users = args.skip_users

    # Does the user want to update from a particular resource?
    from_resources = args.from_resources
    if from_resources == "unset":
        from_resources = None

    # Tell the user to init a particular parser
    if "unset" in args.parsers:
        bot.info("Please specify one or more parsers, one of zenodo, codemeta")
        sys.exit(0)

    if "all" in args.parsers:
        client.init(
            parsers=["zenodo", "allcontrib"],
            repo=args.repo,
            params=extra,
            force=args.force,
            skip_users=skip_users,
            from_resources=from_resources,
        )

    else:
        parsers = [x for x in args.parsers if x != "unset"]
        client.init(
            parsers=parsers,
            repo=args.repo,
            params=extra,
            force=args.force,
            skip_users=skip_users,
            from_resources=from_resources,
        )



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

import sys


def parse_extra(extra):
    """Given a list of extra arguments, parse for known
    """
    known_single = [
        "--zenodo-file",
        "--doi",
        "--allcontrib-type",
        "--allcontrib-file",
        "--codemeta-file",
        "--mailmap-file",
        "--skip-users",
    ]
    known_bool = []

    # Parse extra arguments, all are pairs
    kwargs = {}

    while extra:
        arg = extra.pop(0)
        if arg in known_single:
            if not extra:
                sys.exit(f"Argument {arg} requires a value.")
            kwargs[arg] = extra.pop(0)
        elif arg in known_bool:
            kwargs[arg] = True
    return kwargs



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from tributors.utils.file import write_file, get_tmpfile
from tributors.utils.prompt import choice_prompt
import logging
import os
import requests

bot = logging.getLogger("github")


class OrcidIdentifier:
    """A simple class to retrieve an orcid record, and expose needed fields
    """

    def __init__(self, orcid):
        self.orcid = orcid
        self._record = {}
        self.found = False

    def __str__(self):
        if self.orcid:
            return "[orcid][%s]" % self.orcid
        return "[orcid]"

    def __repr__(self):
        return self.__str__()

    @property
    def record(self):
        """Given an orcid id, retrieve a record for it
        """
        if not self._record and self.orcid:
            self.get_record()
        return self._record

    @property
    def firstName(self):
        return (
            self.record.get("person", {})
            .get("name", {})
            .get("given-names", {})
            .get("value")
        )

    @property
    def affiliation(self):
        """We consider the affiliation the most recent employment (top of the list)
        """
        employer = (
            self.record.get("activities-summary", {})
            .get("employments", {})
            .get("employment-summary", [])
        )
        if employer:
            return employer[0].get("organization", {}).get("name")

    @property
    def lastName(self):
        return (
            self.record.get("person", {})
            .get("name", {})
            .get("family-name", {})
            .get("value")
        )

    def get_record(self):
        if not self.orcid:
            return

        response = requests.get(
            "https://pub.orcid.org/v2.1/%s/record" % self.orcid,
            headers={"Accept": "application/json",},
        )
        if response.status_code != 200:
            return
        self.found = True
        self._record = response.json()


def get_orcid_token():
    """If the user has exported a token, we discover and return it here.
       Otherwise we prompt him or her to open a browser and copy paste a code
       to the calling client. This is currently not used, but kept in case
       we need to add it back.
    """
    orcid_token = os.environ.get("ORCID_TOKEN")
    orcid_id = os.environ.get("ORCID_ID")
    orcid_secret = os.environ.get("ORCID_SECRET")

    if not orcid_token and orcid_id is not None and orcid_secret is not None:
        response = requests.post(
            "https://orcid.org/oauth/token",
            headers={"Accept": "application/json"},
            data={
                "client_id": orcid_id,
                "client_secret": orcid_secret,
                "grant_type": "client_credentials",
                "scope": "/read-public",
            },
        )

        if response.status_code != 200:
            return

        response = response.json()
        orcid_token = response["access_token"]
        orcid_refresh = response["refresh_token"]

        # Write the token to file and direct the user to use it
        tmp_file = get_tmpfile()
        content = "export ORCID_TOKEN=%s\nexport ORCID_REFRESH_TOKEN=%s\n" % (
            orcid_token,
            orcid_refresh,
        )
        write_file(tmp_file, content)
        print(
            f"Orcid token exports written to {tmp_file}. "
            "In the future export these variables for headless usage."
        )

    return orcid_token


def record_search(url, email, interactive=False):
    """Given a url (with a name or email) do a record search looking for an orcid id
    """
    response = requests.get(url, headers={"Accept": "application/json"})
    if response.status_code != 200:
        return

    results = response.json().get("expanded-result", []) or []

    if not results:
        return

    # We found only one matching result
    if len(results) == 1:
        return results[0]["orcid-id"]

    # Only stream results to screen in interactive mode
    if not interactive:
        bot.info(
            f"{email}: found more than 1 result, run with --interactive mode to select."
        )
        return

    # One or more results
    if len(results) > 10:
        bot.warning("Found more than 10 results, will only show top 10.")

    print("\n\n%s\n======================================================" % email)
    for idx, r in enumerate(results):

        # Limit is ten results, count starting at 0
        idx = idx + 1
        if idx > 10:
            break

        record = "  Name: %s, %s\n  Orcid: %s" % (
            r["family-names"],
            r["given-names"],
            r["orcid-id"],
        )
        if r["institution-name"]:
            record = "%s\n  Institutions: %s" % (
                record,
                ", ".join(r["institution-name"]),
            )
        if r["other-name"]:
            record = "%s\n  Other Names: %s" % (record, ", ".join(r["other-name"]))
        if r["email"]:
            record = "%s\n  Email: %s" % (record, r["email"])
        if not interactive:
            print("%s\n" % record)
        else:
            print("[%s]\n%s\n" % (idx, record))

    # If interactive, ask for choice prompt
    if interactive:
        choices = [str(i) for i, _ in enumerate(results)] + ["s", "S", "skip"]
        prefix = "1:%s or s to skip" % ("10" if len(results) > 10 else len(results))
        choice = choice_prompt(
            "Please enter a choice, or s to skip.",
            choices=choices,
            choice_prefix=prefix,
            multiple=False,
        )

        if not choice or choice in ["s", "S", "skip"]:
            return

        # Return the orcid identifier
        return results[int(choice) - 1]["orcid-id"]


def get_orcid(email, name=None, interactive=False):
    """Get an orcid identifier for a given email or name.
    """
    # We must have an email OR name
    if not email and not name:
        return

    # First look for records based on email
    orcid_id = None
    if email:
        url = "https://pub.orcid.org/v3.0/expanded-search?q=email:%s" % email
        orcid_id = record_search(url, email, interactive)

    # Attempt # 2 will use the first and last name
    if name is not None and not orcid_id:

        delim = "," if "," in name else " "
        cleaner = "," if delim == " " else " "

        parts = name.split(delim)

        # No go if only a first or last name
        if len(parts) == 1:
            bot.debug(f"Skipping {name}, first and last are required for search.")
            return orcid_id

        last, first = parts[0].strip(cleaner), " ".join(parts[1:]).strip(cleaner)
        url = "https://pub.orcid.org/v3.0/expanded-search?q=%s+AND+%s" % (first, last)
        orcid_id = record_search(url, name, interactive)

        # Attempt # 3 will try removing the middle name
        if " " in first:
            url = "https://pub.orcid.org/v3.0/expanded-search?q=%s+AND+%s" % (
                first.split(" ")[0].strip(),
                last,
            )
            orcid_id = record_search(url, name, interactive)

        # Last attempt tries full name
        if orcid_id is None:
            url = "https://pub.orcid.org/v3.0/expanded-search?q=%s" % name
            orcid_id = record_search(url, name, interactive)

    return orcid_id



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

import logging
import requests
import os
import sys

from tributors.utils.file import write_json
from .base import ParserBase
from tributors.main.orcid import get_orcid

bot = logging.getLogger("    zenodo")


class ZenodoParser(ParserBase):

    name = "zenodo"

    def __init__(self, filename=None, repo=None, params=None, **kwargs):
        filename = filename or ".zenodo.json"
        super().__init__(filename, repo, params)

    def load_data(self):
        """A shared function to load the zenodo file, if data is not defined
        """
        return self._load_data("--zenodo-file")

    @property
    def email_lookup(self):
        """Return loaded metadata as an email lookup.
        """
        self.load_data()
        return {x["email"]: x for x in self.data.get("creators", []) if "email" in x}

    @property
    def orcid_lookup(self):
        """Return loaded metadata as an orcid lookup.
        """
        self.load_data()
        return {x["orcid"]: x for x in self.data.get("creators", []) if "orcid" in x}

    def init(self, force=False, from_resources=None, save=True):
        """Generate an empty .zenodo.json if it doesn't exist
        """
        from_resources = from_resources or {}
        doi = self.params.get("--doi")

        # Zenodo file defaults to expected .zenodo.json
        zenodo_file = self.params.get("--zenodo-file", self.filename)
        if os.path.exists(zenodo_file) and not force:
            sys.exit("%s exists, set --force to overwrite." % zenodo_file)

        bot.info("Generating %s" % zenodo_file)

        # If a doi is provided, generate
        record = None
        self.data["creators"] = []
        if doi:
            record = get_zenodo_record(doi)
            self.data["creators"] = record["metadata"].get("creators", [])

        self.update_cache(update_lookup=False)

        # Update zenodo file from GitHub logins (default) or other
        self.update_from_logins(from_resources.get("login", []))
        self.update_from_orcids(from_resources.get("orcid", []))
        self.update_from_names(from_resources.get("name", []))
        self.update_from_emails(from_resources.get("email", []))

        # Update final metadata
        metadata = {
            "creators": self.data["creators"],
            "upload_type": "software",
            "keywords": self.repo.topics(),
        }

        # If we have a zenodo record, update it
        if record:
            metadata["upload_type"] = record["metadata"]["resource_type"]["type"]
            metadata["keywords"] = self.repo.topics(record["metadata"]["keywords"])
            metadata["access_right"] = record["metadata"]["access_right"]
            metadata["license"] = record["metadata"]["license"]["id"]

        if save:
            write_json(metadata, zenodo_file)
        return metadata

    def update_from_orcids(self, orcids):
        """Given a list of orcids, update the contributor file from it
        """
        lookup = {x["orcid"]: x for _, x in self.cache.items() if "orcid" in x}
        for orcid in orcids:
            if orcid in self.orcid_lookup:
                continue
            entry = {"orcid": orcid}
            if orcid in lookup:
                for field in ["name", "affiliation", "orcid"]:
                    if field in lookup[orcid] and field not in entry:
                        entry[field] = lookup[orcid][field]
            if entry and entry not in self.data["creators"]:
                self.data["creators"].append(entry)
        return self.data["creators"]

    def update_orcids(self):
        """Zenodo is a special case that has emails and real usernames, so we
           can parse through the existing file and look for orcid identifiers
        """
        interactive = self.params.get("--interactive", False)
        creators = []
        for user in self.data.get("creators", []):
            orcid = user.get("orcid")
            name = user.get("name")
            email = user.get("email")
            if orcid is not None:
                creators.append(user)
                continue
            if email or name:
                orcid = get_orcid(
                    email=email, name=name.strip(), interactive=interactive
                )
                if orcid:
                    user["orcid"] = orcid
            creators.append(user)
        self.data["creators"] = creators

    def update_from_emails(self, emails):
        """Given a list of emails, update the contributor file from it. We also
           look for new orcid ids for emails that don't have them.
        """
        # First loop through emails in the cache
        lookup = {x["email"]: x for _, x in self.cache.items() if "email" in x}
        for email in emails:
            if email in self.email_lookup:
                continue
            entry = {}
            for field in ["name", "affiliation", "orcid"]:
                if email in lookup and field in lookup[email]:
                    entry[field] = lookup[email][field]
            if entry and entry not in self.data["creators"]:
                self.data["creators"].append(entry)
        return self.data["creators"]

    def update_from_logins(self, logins):
        """Given a list of logins, update the zenodo.json from it. We only
           do this on init when we haven't added /updated logins with
           people's actual names.
        """
        # GitHub contributors are the source of truth
        for login in logins:

            # Check against contribution threshold, and not bot
            if not self.include_contributor(login):
                continue

            cache = self.cache.get(login) or {}
            orcid = cache.get("orcid")
            email = cache.get("email")

            # Make sure we don't have already
            if (orcid and orcid in self.orcid_lookup) or (
                email and email in self.email_lookup
            ):
                continue

            entry = {"name": cache.get("name") or login}
            if login in self.cache:
                for field in ["name", "affiliation", "orcid"]:
                    if field in self.cache[login]:
                        entry[field] = self.cache[login][field]
            if "orcid" in cache and "orcid" not in entry:
                entry["orcid"] = cache["orcid"]
            if "affiliation" in cache and "affiliation" not in entry:
                entry["affilitation"] = cache["affiliation"]

            # Don't add duplicates
            if entry not in self.data["creators"]:
                self.data["creators"].append(entry)
        return self.data["creators"]

    def update(self, thresh=1, from_resources=None, save=True):
        """Given an existing .zenodo.json file, update it with contributors
           from an allcontributors file.
        """
        from_resources = from_resources or {}
        self.thresh = thresh
        self.load_data()
        bot.info("Updating %s" % self.filename)

        self.update_cache()

        # Here we can only reasonable update from orcids (not logins)
        self.update_orcids()
        self.update_from_emails(from_resources.get("email", []))
        self.update_from_orcids(from_resources.get("orcid", []))
        self.update_from_names(from_resources.get("names", []))
        if save:
            write_json(self.data, self.filename)
        return self.data

    def update_lookup(self):
        """Each client optionally has it's own function to update the cache.
            In the case of zenodo, we aren't necessarily aware of GitHub
            login (the current mapping mechanism) so we cannot update the
            cache yet. When orcid is added this might be an option.
        """
        self.load_data()
        bot.info(f"Updating .tributors cache from {self.filename}")

        # We have to update based on orcid
        lookup = {}
        for entry in self.data.get("creators", []):
            if "orcid" in entry:
                lookup[entry["orcid"]] = entry

        # Now loop through cache
        for login, cache in self.cache.items():
            if "orcid" in cache and cache["orcid"] in lookup:
                for field in ["name", "affiliation"]:
                    if field in lookup[cache["orcid"]] and field not in cache:
                        value = lookup[cache]["orcid"][field]
                        bot.info(f"   Updating {login} with {field}: {value}")
                        cache[field] = value


def get_zenodo_record(doi):
    """Given a doi, retrieve a record using the Zenodo API
    """
    # Get the record number from the doi
    record = doi.split("/")[-1].replace("zenodo.", "")
    token = os.environ.get("ZENODO_TOKEN")
    if token:
        response = requests.get(
            "https://zenodo.org/api/records/%s" % record, json={"access_token": token},
        )
    else:
        response = requests.get("https://zenodo.org/api/records/%s" % record)

    # Successful query!
    if response.status_code != 200:
        sys.exit(
            "Response %s: %s, cannot retrieve zenodo %s."
            % (response.status_code, response.reason, doi)
        )
    return response.json()



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

import logging

from tributors.utils.file import write_json
from .base import ParserBase

bot = logging.getLogger("  codemeta")


class CodeMetaParser(ParserBase):

    name = "codemeta"

    def __init__(self, filename=None, repo=None, params=None, **kwargs):
        filename = filename or "codemeta.json"
        self.data = {}
        super().__init__(filename, repo, params)

    def init(self, force=False):
        """Codemeta already has many good generators.
        """
        print(
            "Codemeta provides several tools to generate this for you: https://codemeta.github.io/tools/"
        )

    def load_data(self):
        """If not yet loaded, load data into client
        """
        return self._load_data("--codemeta-file")

    def update(self, thresh=1, from_resources=None, save=True):
        """Given an existing .zenodo.json file, update it with contributors
           from an allcontributors file.
        """
        self.thresh = thresh
        self.load_data()

        bot.info("Updating %s" % self.filename)

        # Read in contributors, and update cache (also runs update_lookup)
        self.lookup = self.data.get("contributor", [])
        self.update_cache()

        # Get fields from repo
        self.update_metadata()

        self.update_from_logins(from_resources.get("login", []))
        self.update_from_orcids(from_resources.get("orcid", []))
        self.update_from_names(from_resources.get("name", []))
        self.update_from_emails(from_resources.get("email", []))

        self.data["contributor"] = self.lookup
        if save:
            write_json(self.data, self.filename)
        return self.data

    def update_metadata(self):
        """Update codemeta metadata from the repository, if we can.
        """
        self.data["keywords"] = self.repo.topics(self.data["keywords"])
        self.data["description"] = self.data.get("description") or self.repo.description
        self.data["codeRepository"] = (
            self.data.get("codeRepository") or self.repo.html_url
        )
        self.data["name"] = self.data.get("name") or self.repo.name
        self.data["issueTracker"] = (
            self.data.get("issueTracker") or self.repo.issues_url
        )
        self.data["license"] = self.data.get("license") or self.repo.license

    def update_from_emails(self, emails):
        """Update codemeta entries from emails
        """
        # Now add contributors using cache (new GitHub contributors) with known email or orcid that isn't present
        for email in emails:
            if email not in self.email_lookup:
                bot.info(f"   Updating with new added email: {email}")
                entry = {"@type": "Person", "email": email}
                self.lookup.append(entry)

    def update_from_logins(self, logins):
        """Update codemeta entries from GitHub logins
        """
        # Now add contributors using cache (new GitHub contributors) with known email or orcid that isn't present
        for login in logins:

            # Check against contribution threshold, and not bot
            if not self.include_contributor(login):
                continue

            cache = self.cache.get(login) or {}
            email = cache.get("email")
            orcid = cache.get("orcid")

            # We can only add completely new entries that don't already exist
            if (email != None or orcid != None) and (
                email not in self.email_lookup and orcid not in self.orcid_lookup
            ):
                bot.info(f"   Updating {login}")
                parts = (cache.get("name") or login).split(" ")
                entry = {"@type": "Person", "givenName": parts[0]}

                # Add the last name if it's defined
                if len(parts) > 1:
                    entry["familyName"] = " ".join(parts[1:])

                if email != None:
                    entry["email"] = email
                if orcid != None:
                    entry["@id"] = "https://orcid.org/%s" % orcid
                self.lookup.append(entry)

    @property
    def email_lookup(self):
        """Return loaded metadata as an email lookup
        """
        if not hasattr(self, "_email_lookup"):
            self._email_lookup = {}
            self.load_data()
        for entry in self.data.get("contributor", []):
            if "email" in entry:
                self._email_lookup[entry["email"]] = entry
        return self._email_lookup

    @property
    def orcid_lookup(self):
        """Return loaded metadata as an orcid lookup
        """
        if not hasattr(self, "_orcid_lookup"):
            self._orcid_lookup = {}
            self.load_data()
            for entry in self.data.get("contributor", []):
                if "@id" in entry:
                    # Orcid represented as full URL but we just want id
                    orcid = entry["@id"].split("/")[-1]
                    self._orcid_lookup[orcid] = entry
        return self._orcid_lookup

    def update_lookup(self):
        """We can only keep track of users here based on email addresses or
           orcid, so we can only update the cache for existing users.
        """
        bot.info(f"Updating .tributors cache from {self.filename}")

        # Find matches based on orcid and/or email
        for login, cache in self.cache.items():
            orcid = cache.get("orcid")
            email = cache.get("email")

            # Case 1: double match (unlikely but possible)
            entry = None
            if email in self.email_lookup and orcid in self.orcid_lookup:

                # If they don't point to the same entry, stop
                if self.email_lookup[email] != self.orcid_lookup[orcid]:
                    bot.warning(
                        "Found email {email} and orcid {orcid} in cache from different entries, skipping."
                    )
                else:
                    entry = self.email_lookup[email]

            # Case 2: We have a matching orcid
            elif orcid in self.orcid_lookup:
                entry = self.orcid_lookup[orcid]

            # Case 2: We have a matching email
            elif email in self.email_lookup:
                entry = self.email_lookup[email]

            # If we have a match (entry is defined) use it to update the record
            if entry is not None:

                # Update the name
                if (
                    "givenName" in entry
                    and "familyName" in entry
                    and "name" not in cache
                ):
                    cache["name"] = "%s %s" % (entry["givenName"], entry["familyName"])
                    bot.info(f"   Updating {login} with name: {cache['name']}")

                # Update the email
                if "email" in entry and "emali" not in cache:
                    cache["email"] = entry["email"]
                    bot.info(f"   Updating {login} with email: {cache['email']}")

                # Update the orcid id
                if "@id" in entry and "orcid" not in cache:
                    cache["orcid"] = entry["@id"].split("/")[-1]
                    bot.info(f"   Updating {login} with orcid: {cache['orcid']}")



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""


enter_input = getattr(__builtins__, "raw_input", input)


def request_input():
    """Wait for the user to input some string, optionally with multiple lines.
    """
    lines = []

    # The message can be multiple lines
    while True:
        try:
            line = enter_input()
        except EOFError:
            break
        if line:
            lines.append(line)

    return "\n".join(lines)


def choice_prompt(prompt, choices, choice_prefix=None, multiple=False):
    """Ask the user for a prompt, and only return when one of the requested
       options is provided.
       Parameters
       ==========
       prompt: the prompt to ask the user
       choices: a list of choices that are valid.   
       multiple: allow multiple responses (separated by spaces)
    """
    choice = None
    print(prompt)

    # Support for Python 2 (raw_input)
    get_input = getattr(__builtins__, "raw_input", input)

    if not choice_prefix:
        choice_prefix = "/".join(choices)
    message = "[%s] : " % (choice_prefix)

    while choice not in choices:
        choice = get_input(message).strip()

        # If multiple allowed, add selection to choices if includes all vaid
        if multiple is True:
            contenders = choice.strip().split(" ")
            if all(x in choices for x in contenders):
                choices.append(choice)
        message = "Please enter a valid option in [%s]" % choice_prefix
    return choice



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

__version__ = "0.0.19"
AUTHOR = "Vanessa Sochat"
AUTHOR_EMAIL = "vsochat@stanford.edu"
NAME = "tributors"
PACKAGE_URL = "http://www.github.com/con/tributors"
KEYWORDS = "contributions, tributes"
DESCRIPTION = "give tribute to your repository contributors."
LICENSE = "LICENSE"

################################################################################
# Global requirements


INSTALL_REQUIRES = (("requests", {"min_version": "2.23.0"}),)
CODEMETA_REQUIRES = (("codemetapy", {"min_version": "0.3.2"}),)
TESTS_REQUIRES = (("pytest", {"min_version": "4.6.2"}),)

ALL_REQUIRES = INSTALL_REQUIRES + CODEMETA_REQUIRES



#!/usr/bin/env python3

"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from tributors.logger import LOG_LEVEL, LOG_LEVELS

import tributors
import argparse
import logging
import sys


def get_parser():
    parser = argparse.ArgumentParser(description="Tributors Python Metadata Parser")

    parser.add_argument(
        "--log-level",
        dest="log_level",
        choices=LOG_LEVELS,
        default=LOG_LEVEL,
        help="Customize logging level for a2z parser.",
    )

    parser.add_argument(
        "--version",
        dest="version",
        help="suppress additional output.",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--skip-cache",
        dest="skip_cache",
        help="Skip loading and saving a .tributors cache file",
        default=False,
        action="store_true",
    )

    description = "actions for tributors"
    subparsers = parser.add_subparsers(
        help="tributors actions",
        title="actions",
        description=description,
        dest="command",
    )

    init = subparsers.add_parser(
        "init", help="Initialize an empty .all-contributorsrc."
    )
    init.add_argument(
        "--force",
        dest="force",
        help="If exists, overwrite existing .all-contributorsrc",
        default=False,
        action="store_true",
    )

    # print version and exit
    subparsers.add_parser("version", help="show software version")

    # Update the .tributors lookup
    update_lookup = subparsers.add_parser(
        "update-lookup", help="Update shared .tributors metadata file",
    )
    update_lookup.add_argument(
        "files",
        help="One or more files to use for update.",
        nargs="*",
        default="unset",
        choices=["mailmap", "allcontrib", "codemeta", "zenodo", "unset", "github"],
    )

    # Update an existing contributors file
    update = subparsers.add_parser("update", help="Update existing all-contributorsrc",)
    update.add_argument(
        "--thresh",
        dest="thresh",
        help="Minimum number of contributions required to add user as contributor",
        default=1,
        type=int,
    )

    for command in [update, init]:
        command.add_argument(
            "--interactive",
            dest="interactive",
            help="Allow interactive selection of orcid ids.",
            default=False,
            action="store_true",
        )
        command.add_argument(
            "parsers",
            help="Metadata file parsers to update or initialize.",
            nargs="*",
            default="unset",
            choices=["zenodo", "allcontrib", "codemeta", "all", "unset"],
        )
        command.add_argument(
            "--repo", help="The repository URI, if not exported to GITHUB_REPOSITORY",
        )
        command.add_argument(
            "--skip-users",
            help="GitHub aliases to skip adding to .tributors file",
            dest="skip_users",
            nargs="*",
            default="unset",
        )
        command.add_argument(
            "--from",
            dest="from_resources",
            help="Update records FROM one resource TO another using unique ids (email, orcid, GitHub login)",
            default="unset",
            choices=[
                "unset",
                "tributors",
                "codemeta",
                "zenodo",
                "github",
                "allcontrib",
                "mailmap",
            ],
            nargs="*",
        )

    return parser


def main():
    """main entrypoint for tributors
    """

    parser = get_parser()

    def help(return_code=0):
        print("\nTributors Python")
        parser.print_help()
        sys.exit(return_code)

    # If the user didn't provide any arguments, show the full help
    if len(sys.argv) == 1:
        help()

    # Each parser takes extra arguments
    args, extra = parser.parse_known_args()

    # Set the logging level
    logging.basicConfig(level=getattr(logging, args.log_level))
    bot = logging.getLogger("tributors")
    bot.setLevel(getattr(logging, args.log_level))

    # Show the version and exit
    if args.command == "version" or args.version:
        print(tributors.__version__)
        sys.exit(0)

    # Initialize or update one or more configuration files
    if args.command == "init":
        from .init import main
    elif args.command == "update":
        from .update import main
    elif args.command == "update-lookup":
        from .lookup import main
    else:
        help()

    main(args, extra)


if __name__ == "__main__":
    main()



"""

Copyright (C) 2020 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from tributors.main import TributorsClient
from .utils import parse_extra
import os
import sys


def main(args, extra):

    client = TributorsClient(skip_cache=args.skip_cache)

    # Parse extra arguments
    extra = parse_extra(extra)

    # Start with user provided parsers
    resources = args.files

    # If unset, try to detect files
    if "unset" in resources:
        lookup = {
            "allcontrib": extra.get("--allcontrib-file", ".all-contributorsrc"),
            "zenodo": extra.get("--zenodo-file", ".zenodo.json"),
            "codemeta": extra.get("--codemeta-file", "codemeta.json"),
            "mailmap": extra.get("--mailmap-file", ".mailmap"),
        }

        resources = []
        for resource, filename in lookup.items():
            if os.path.exists(filename):
                resources.append(resource)

        # Exit if no parsers auto-detected
        if not resources:
            sys.exit("No resources auto-detected. Specify one or more instead?")

    if "all" in resources:
        client.update_resource(
            resources=["zenodo", "allcontrib", "codemeta", "mailmap", "github"],
            params=extra,
        )

    else:
        client.update_resource(resources=resources, params=extra)
